searchNodes=[{"doc":"Job pools for Elixir OTP applications, written in Elixir. Use OddJob when you want to limit concurrency of background processing in your Elixir app. One possible use case is forcing backpressure on calls to external APIs. Installation The package can be installed by adding odd_job to your list of dependencies in mix.exs : def deps do [ { :odd_job , &quot;~&gt; 0.3.2&quot; } ] end Getting started OddJob automatically starts up a supervised job pool of 5 workers out of the box with no extra configuration. The default name of this job pool is :job , and it can be sent work in the following way: OddJob . perform ( :job , fn -&gt; do_some_work ( ) end ) The default pool can be customized in your config if you want to change the name or pool size: config :odd_job , default_pool : :work , pool_size : 10 # this changes the size of all pools in your application, defaults to 5 You can also add extra pools to be supervised by the OddJob application supervision tree: config :odd_job , extra_pools : [ :email , :external_app ] If you don't want OddJob to supervise any pools for you (including the default :job pool) then pass false to the :default_pool config key: config :odd_job , default_pool : false Supervising job pools To supervise your own job pools you can add a tuple in the form of {OddJob, name} (where name is an atom) directly to the top level of your application's supervision tree or any other list of child specs for a supervisor: defmodule MyApp.Application do use Application def start ( _type , _args ) do children = [ { OddJob , :email } , { OddJob , :external_app } ] opts = [ strategy : :one_for_one , name : MyApp.Supervisor ] Supervisor . start_link ( children , opts ) end end The tuple {OddJob, :email} will return a child spec for a supervisor that will start and supervise the :email pool. The second element of the tuple can be any atom that you want to use as a unique name for the pool. You can supervise as many pools as you want, as long as they have unique name s. All of the aforementioned config options can be combined. You can have a default pool (with an optional custom name), extra pools in the OddJob supervision tree, and pools to be supervised by your own application. Basic usage and async/await A job pool can be sent jobs by passing its unique name and an anonymous function to one of the OddJob module's perform functions: job = OddJob . async_perform ( :external_app , fn -&gt; get_data ( user ) end ) # do something else data = OddJob . await ( job ) OddJob . perform ( :email , fn -&gt; send_email ( user , data ) end ) If a worker in the pool is available then the job will be performed right away. If all of the workers are already assigned to other jobs then the new job will be added to a FIFO queue. Jobs in the queue are performed as workers become available. Use perform/2 for immediate fire and forget jobs where you don't care about the results or if it succeeds. async_perform/2 and await/1 follow the async/await pattern in the Task module, and are useful when you need to retrieve the results and you care about success or failure. Similarly to Task.async/1 , async jobs will be linked and monitored by the caller (in this case, through a proxy). If either the caller or the job crash or exit, the other will crash or exit with the same reason. Scheduled jobs Jobs can be scheduled for later execution with perform_after/3 and perform_at/3 : OddJob . perform_after ( 1_000_000 , :job , fn -&gt; clean_database ( ) end ) # accepts a timer in milliseconds time = ~T[03:00:00.000000] OddJob . perform_at ( time , :job , fn -&gt; verify_work_is_done ( ) end ) # accepts a valid Time or DateTime struct The scheduling functions return a unique timer reference which can be read with Process.read_timer/1 and cancelled with OddJob.cancel_timer/1 , which will cancel execution of the job itself and cause the scheduler process to exit. When the timer is up the job will be sent to the pool and can no longer be aborted. ref = OddJob . perform_after ( 5000 , :job , fn -&gt; :will_be_canceled end ) # somewhere else in your code if some_condition ( ) do OddJob . cancel_timer ( ref ) end Note that there is no guarantee that a scheduled job will be performed right away when the timer runs out. Like all jobs it is sent to the pool and if all workers are busy at that time then the job enters the queue to be performed when a worker is available. License MIT - Copyright (c) 2022 M. Simon Borg","ref":"OddJob.html","title":"OddJob","type":"module"},{"doc":"Performs an async job that can be awaited on for the result. Functions like Task.async/1 and Task.await/2 . Examples iex&gt; job = OddJob . async_perform ( :job , fn -&gt; :math . exp ( 100 ) end ) iex&gt; OddJob . await ( job ) 2.6881171418161356e43","ref":"OddJob.html#async_perform/2","title":"OddJob.async_perform/2","type":"function"},{"doc":"Awaits on an async job and returns the results. Examples iex&gt; OddJob . async_perform ( :job , fn -&gt; :math . log ( 2.6881171418161356e43 ) end ) ...&gt; |&gt; OddJob . await ( ) 100.0","ref":"OddJob.html#await/2","title":"OddJob.await/2","type":"function"},{"doc":"Awaits replies form multiple async jobs and returns them in a list. This function receives a list of jobs and waits for their replies in the given time interval. It returns a list of the results, in the same order as the jobs supplied in the jobs input argument. If any of the job worker processes dies, the caller process will exit with the same reason as that worker. A timeout, in milliseconds or :infinity , can be given with a default value of 5000 . If the timeout is exceeded, then the caller process will exit. Any worker processes that are linked to the caller process (which is the case when a job is started with async_perform/2 ) will also exit. This function assumes the jobs' monitors are still active or the monitor's :DOWN message is in the message queue. If any jobs have been demonitored, or the message already received, this function will wait for the duration of the timeout. Examples iex&gt; job1 = OddJob . async_perform ( :job , fn -&gt; 2 * * 2 end ) iex&gt; job2 = OddJob . async_perform ( :job , fn -&gt; 3 * * 2 end ) iex&gt; [ job1 , job2 ] |&gt; OddJob . await_many ( ) [ 4 , 9 ]","ref":"OddJob.html#await_many/2","title":"OddJob.await_many/2","type":"function"},{"doc":"Cancels a scheduled job. timer_ref is the unique reference returned by perform_at/3 or perform_after/3 . This function returns the number of milliseconds left in the timer when cancelled, or false if the timer already expired. If the return is false you can assume that the job has already been sent to the pool for execution. NOTE: Cancelling the timer with this function ensures that the job is never executed and that the scheduler process is exited and not left &quot;hanging&quot;. Using Process.cancel_timer/1 will also cancel execution, but may leave hanging processes. A hanging scheduler process will eventually timeout, but not until one second after the expiration of the original timer. Examples iex&gt; ref = OddJob . perform_after ( 500 , :work , fn -&gt; :never end ) iex&gt; time = OddJob . cancel_timer ( ref ) iex&gt; is_integer ( time ) true iex&gt; ref = OddJob . perform_after ( 10 , :work , fn -&gt; :never end ) iex&gt; Process . sleep ( 11 ) iex&gt; OddJob . cancel_timer ( ref ) false","ref":"OddJob.html#cancel_timer/1","title":"OddJob.cancel_timer/1","type":"function"},{"doc":"Performs a fire and forget job. Examples iex&gt; parent = self ( ) iex&gt; :ok = OddJob . perform ( :job , fn -&gt; send ( parent , :hello ) end ) iex&gt; receive do ...&gt; msg -&gt; msg ...&gt; end :hello","ref":"OddJob.html#perform/2","title":"OddJob.perform/2","type":"function"},{"doc":"Sends a job to the pool after the given timer has elapsed. timer is an integer that indicates the number of milliseconds that should elapse before the job is sent to the pool. The timed message is executed under a separate supervised process, so if the caller crashes the job will still be performed. A timer reference is returned, which can be read with Process.read_timer/1 or cancelled with OddJob.cancel_timer/1 . Examples timer_ref = OddJob . perform_after ( 5000 , :job , fn -&gt; deferred_job ( ) end ) Process . read_timer ( timer_ref ) #=&gt; 2836 # time remaining before job is sent to the pool OddJob . cancel_timer ( timer_ref ) #=&gt; 1175 # job has been cancelled timer_ref = OddJob . perform_after ( 5000 , :job , fn -&gt; deferred_job ( ) end ) Process . sleep ( 6000 ) OddJob . cancel_timer ( timer_ref ) #=&gt; false # too much time has passed to cancel the job","ref":"OddJob.html#perform_after/3","title":"OddJob.perform_after/3","type":"function"},{"doc":"Sends a job to the pool at the given time . time can be a Time or a DateTime struct. If a Time struct is received, then the job will be done the next time the clock strikes the given time. The timer is executed under a separate supervised process, so if the caller crashes the job will still be performed. A timer reference is returned, which can be read with Process.read_timer/1 or canceled with OddJob.cancel_timer/1 . Examples time = Time . utc_now ( ) |&gt; Time . add ( 600 , :second ) OddJob . perform_at ( time , :job , fn -&gt; scheduled_job ( ) end )","ref":"OddJob.html#perform_at/3","title":"OddJob.perform_at/3","type":"function"},{"doc":"A macro for creating jobs with an expressive DSL. perform_this/2 works like perform/2 except it accepts a do block instead of an anonymous function. Examples You must import or require OddJob to use macros: import OddJob perform_this :work do some_work ( ) some_other_work ( ) end perform_this :work , do : something_hard ( )","ref":"OddJob.html#perform_this/2","title":"OddJob.perform_this/2","type":"macro"},{"doc":"A macro for creating jobs with an expressive DSL. perform_this/3 accepts a single configuration option as the second argument that will control execution of the job. The available options prvode the functionality of async_perform/2 , perform_at/3 , and perform_after/3 . Options :async - Passing the atom :async as the second argument before the do block creates an async job that can be awaited on. See async_perform/2 . at: time - Use this option to schedule the job for a specific time in the future. time must be a valid Time or DateTime struct. See perform_at/3 . after: timer - Use this option to schedule the job to perform after the given timer has elapsed. timer must be in milliseconds. See perform_after/3 . Examples import OddJob time = ~T[03:00:00.000000] perform_this :work , at : time do scheduled_work ( ) end perform_this :work , after : 5000 , do : something_important ( ) perform_this :work , :async do get_data ( ) end |&gt; await ( ) iex&gt; ( perform_this :work , :async , do : 10 * * 2 ) |&gt; await ( ) 100","ref":"OddJob.html#perform_this/3","title":"OddJob.perform_this/3","type":"macro"},{"doc":"Returns the pid and state of the job pool . Examples iex&gt; { pid , % OddJob.Pool { id : id } } = OddJob . pool ( :job ) iex&gt; is_pid ( pid ) true iex&gt; id :job_pool","ref":"OddJob.html#pool/1","title":"OddJob.pool/1","type":"function"},{"doc":"Returns the ID of the job pool . Examples iex&gt; OddJob . pool_id ( :job ) :job_pool","ref":"OddJob.html#pool_id/1","title":"OddJob.pool_id/1","type":"function"},{"doc":"Returns the pid of the job pool 's supervisor. There is no guarantee that the process will still be alive after the results are returned, as it could exit or be killed or restarted at any time. Use supervisor_id/1 to obtain the persistent ID of the supervisor. Examples OddJob . supervisor ( :job ) #=&gt; #PID&lt;0.239.0&gt;","ref":"OddJob.html#supervisor/1","title":"OddJob.supervisor/1","type":"function"},{"doc":"Returns the ID of the job pool 's supervisor. Examples iex&gt; OddJob . supervisor_id ( :job ) :job_pool_sup","ref":"OddJob.html#supervisor_id/1","title":"OddJob.supervisor_id/1","type":"function"},{"doc":"Returns a list of pid s for the specified worker pool. There is no guarantee that the processes will still be alive after the results are returned, as they could exit or be killed at any time. Examples OddJob . workers ( :job ) #=&gt; [#PID&lt;0.105.0&gt;, #PID&lt;0.106.0&gt;, #PID&lt;0.107.0&gt;, #PID&lt;0.108.0&gt;, #PID&lt;0.109.0&gt;]","ref":"OddJob.html#workers/1","title":"OddJob.workers/1","type":"function"},{"doc":"","ref":"OddJob.html#t:child_spec/0","title":"OddJob.child_spec/0","type":"type"},{"doc":"","ref":"OddJob.html#t:job/0","title":"OddJob.job/0","type":"type"},{"doc":"","ref":"OddJob.html#t:pool/0","title":"OddJob.pool/0","type":"type"},{"doc":"The OddJob.Job struct holds all of the useful information about a job.","ref":"OddJob.Job.html","title":"OddJob.Job","type":"module"},{"doc":"The OddJob.Job struct is the datatype that is passed between processes charged with performing the job. It holds all of the data that is necessary to link, monitor, perform work, and return results to the caller. The job struct is only returned to the caller when using the async/await pattern. When the caller receives the struct after calling OddJob.async_perform/2 the :results field is always nil , even though the work could conceivably already be done. This is because the results are not waited on at the time the struct is created. The results are only known when passing the job to OddJob.await/2 or matching on the {ref, results} message. :function is the anonymous function that will be performed by the worker :results is the term that is returned by function . This is only used internally by the processes performing the work. :async is an boolean identifying if the job's results can be awaited on :ref is the unique monitor reference of the job :owner is the pid of the calling process, i.e. self() :proxy is the pid of the proxy server that creates the job and routes the results. The owner links and monitors the proxy , while the proxy links and monitors the worker. Exit messages and failures cascade up to the owner . The worker sends results back to the proxy , which then sends them to the owner before exiting with reason :normal .","ref":"OddJob.Job.html#t:t/0","title":"OddJob.Job.t/0","type":"type"},{"doc":"The job pool that manages the assignments given to the pool workers.","ref":"OddJob.Pool.html","title":"OddJob.Pool","type":"module"},{"doc":"","ref":"OddJob.Pool.html#t:job/0","title":"OddJob.Pool.job/0","type":"type"},{"doc":"The OddJob.Pool struct holds the state of the job pool. :id is an atom representing the registered name of the pool process :pool is an atom representing the name of the job pool :workers is a list of the active worker pid s, whether they are busy working or not :assigned is a list of the worker pid s that are currently assigned to a job :jobs is a list of OddJob.Job structs representing the jobs that are queued to be performed when workers are available","ref":"OddJob.Pool.html#t:t/0","title":"OddJob.Pool.t/0","type":"type"}]